/**
 * \file AlphaLCD.cpp
 * \brief AlphaLCD, allows the LCD to be operated via a shift register
 *
 * \page LCD Alphanumeric Display Controller
 * This class controls the hardware LCD Alphanumeric display. The details on the usage
 * and hardware features are described in the the Display datasheet component data sheet\n
 This library is a variant derived by the original Arduino ShiftLCD library
 with some improvements and more parametrization. For better readability and
 developers customization and usage a doxygen-based comment set has been added.
 *
 * \section copy_sec Credits, Copyright and License
 * \date January 2015
 * \version 2.0
 * \author Created by \a Enrico \a Miglino for Balearic Dynamics Spain\n
 * Documentation revision by \a Maria \a Antonietta \a Ricagno \n
 * Info and contacts: balearicdynamics@gmail.com - http://www.balearicdynamics.com
 *
 * \note This software is licensed as open source under the GNU license.
*/

#include "AlphaLCD.h"

#include <stdio.h>
#include <string.h>
#include <inttypes.h>
#include "Arduino.h"

/**
 * Constructor with no parameters, to create initial class instances
 */
AlphaLCD::AlphaLCD() {

}

/**
 * Constructor with the hardware initialization parameters (pin numbers)
 *
 * \brief parameters depends on the phisycal connection of the LCD-Alpha component.
 */
AlphaLCD::AlphaLCD(uint8_t dp, uint8_t cp, uint8_t lp)
{
	init (dp, cp, lp);
}

/**
 * \brief Hardware initialization
 *
 Initializes the hardware parameters for the LCD connection. The LCD mode
 is set to 4 bit by default due the hardware circuit settings
 *
 * \param int Data Pin number
 * \param int Clock Pin number
 * \param int Latch Pin number
 */
void AlphaLCD::init(uint8_t dp, uint8_t cp, uint8_t lp)
{
	_data_pin = dp;
	_clock_pin = cp;
	_latch_pin = lp;

	_backlight = LCD_BL_PIN;

	pinMode(_data_pin, OUTPUT);
	pinMode(_clock_pin, OUTPUT);
	pinMode(_latch_pin, OUTPUT);

    _displayfunction = LCD_4BITMODE | LCD_1LINE | LCD_5x8DOTS;

	// Display size (characters and lines). By default it is set to 16 characters x 2 lines
	// but should be set with the right number of lines and columns,
	// accordingly with the hardware specifications
  begin(16, 2);
}

/**
 * Start the LCD modes and initializes the software configuration parameters.
 * \note The method avoid displays with more than 2 lines
 *
 * \params int Number of columns (hardware settings: 16 characters)
 * \params int number of lines (hardware settings: 2 lines)
 */
void AlphaLCD::begin(uint8_t cols, uint8_t lines, uint8_t dotsize) {
  if (lines > 1) {
    _displayfunction |= LCD_2LINE;
  }

  _numlines = lines;
  _currline = 0;

  // for some 1 line displays you can select a 10 pixel height font
  if ((dotsize != 0) && (lines == 1)) {
    _displayfunction |= LCD_5x10DOTS;
  }

  // We need at least 40ms after power rises before sending commands.
  delayMicroseconds(50000);
  // clear the shift register by sending 16 0's to it (twice)
	shiftOut(_data_pin, _clock_pin, LSBFIRST, B00000000);
	shiftOut(_data_pin, _clock_pin, LSBFIRST, B00000000);

  //put the LCD into 4 bit or 8 bit mode
  if (! (_displayfunction & LCD_8BITMODE)) {

    // we start in 8bit mode, try to set 4 bit mode
    write4bits(0x03, LOW);
    delayMicroseconds(4500);

    // second try
    write4bits(0x03, LOW);
    delayMicroseconds(4500);

    // third go!
    write4bits(0x03, LOW);
    delayMicroseconds(150);

    // set to 8-bit interface
    write4bits(0x02, LOW);
  } else {
    // Send function set command sequence
    command(LCD_FUNCTIONSET | _displayfunction);
    delayMicroseconds(4500);

    // second try
    command(LCD_FUNCTIONSET | _displayfunction);
    delayMicroseconds(150);

    // third go
    command(LCD_FUNCTIONSET | _displayfunction);
  }

  // finally, set # lines, font size, etc.
  command(LCD_FUNCTIONSET | _displayfunction);

  // turn the display on with no cursor or blinking by default
  _displaycontrol = LCD_DISPLAYON | LCD_CURSOROFF | LCD_BLINKOFF;
  display();

  // clear it off
  clear();

  // Initialize to default text direction (for romance languages)
  _displaymode = LCD_ENTRYLEFT | LCD_ENTRYSHIFTDECREMENT;
  // set the entry mode
  command(LCD_ENTRYMODESET | _displaymode);

}

/**
 * Clear the display content and set the cursor to the position (0,0)
 */
void AlphaLCD::clear()
{
  command(LCD_CLEARDISPLAY);
  delayMicroseconds(2000);
}

/**
 * Set the cursor to the position (0,0)
 */
void AlphaLCD::home()
{
  command(LCD_RETURNHOME);
  delayMicroseconds(2000);
}

/**
 * Set the cursor to the requested position. The method take care avoiding out of bound
 * row and column values.
 *
 * \note The row and column are zero-based offset.
 *
 * \param int column
 * \param int row
 */
void AlphaLCD::setCursor(uint8_t col, uint8_t row)
{
  int row_offsets[] = { 0x00, 0x40, 0x14, 0x54 };
  if ( row > _numlines ) {
    row = _numlines-1;
  }

  command(LCD_SETDDRAMADDR | (col + row_offsets[row]));
}

/**
  *Turn off the display
 */
void AlphaLCD::noDisplay() {
  _displaycontrol &= ~LCD_DISPLAYON;
  command(LCD_DISPLAYCONTROL | _displaycontrol);
}

/**
 * Turn on the display
 */
void AlphaLCD::display() {
  _displaycontrol |= LCD_DISPLAYON;
  command(LCD_DISPLAYCONTROL | _displaycontrol);
}

/**
 * Turn off the underline cursor
 */
void AlphaLCD::noCursor() {
  _displaycontrol &= ~LCD_CURSORON;
  command(LCD_DISPLAYCONTROL | _displaycontrol);
}

/**
 * Turn on the underline cursor
 */
void AlphaLCD::cursor() {
  _displaycontrol |= LCD_CURSORON;
  command(LCD_DISPLAYCONTROL | _displaycontrol);
}

/**
 * Turn off the blinking cursor
 */
void AlphaLCD::noBlink() {
  _displaycontrol &= ~LCD_BLINKON;
  command(LCD_DISPLAYCONTROL | _displaycontrol);
}

/**
 * Turn off the blinking cursor
 */
void AlphaLCD::blink() {
  _displaycontrol |= LCD_BLINKON;
  command(LCD_DISPLAYCONTROL | _displaycontrol);
}

/**
 * Scroll the display to the left by one position. The method does not
 * change the ram content.
 */
void AlphaLCD::scrollDisplayLeft(void) {
  command(LCD_CURSORSHIFT | LCD_DISPLAYMOVE | LCD_MOVELEFT);
}

/**
 * Scroll the display to the right by one position. The method does not
 * change the ram content.
 */
void AlphaLCD::scrollDisplayRight(void) {
  command(LCD_CURSORSHIFT | LCD_DISPLAYMOVE | LCD_MOVERIGHT);
}

/**
 * Flow the text from left to right
 */
void AlphaLCD::leftToRight(void) {
  _displaymode |= LCD_ENTRYLEFT;
  command(LCD_ENTRYMODESET | _displaymode);
}

/**
 * Flow the text from right to left
 */
void AlphaLCD::rightToLeft(void) {
  _displaymode &= ~LCD_ENTRYLEFT;
  command(LCD_ENTRYMODESET | _displaymode);
}

/**
 * Enable the automatic horizontal scrolling of the text. As the scrolling is done
 * on the entire display content, the resulting effect is like the right justification
 */
void AlphaLCD::autoscroll(void) {
  _displaymode |= LCD_ENTRYSHIFTINCREMENT;
  command(LCD_ENTRYMODESET | _displaymode);
}

/**
 * Disable the automatic horizontal scrolling of the text.
 */
void AlphaLCD::noAutoscroll(void) {
  _displaymode &= ~LCD_ENTRYSHIFTINCREMENT;
  command(LCD_ENTRYMODESET | _displaymode);
}

//! Helper method to set on/off the display
//! \param bool the flag setting
inline void AlphaLCD::isDisplay(bool set) {
	set ? display() : noDisplay();
}

//! Helper method to set on/off the cursor visibility
//! \param bool the flag setting
inline void AlphaLCD::isCursor(bool set) {
	set ? cursor() : noCursor();
}

//! Helper method to set on/off the blinking cursor
//! \param bool the flag setting
inline void AlphaLCD::isBlinking(bool set) {
	set ? blink() : noBlink();
}

//! Helper method to set on/off the right-to-left writing direction
//! \param bool the flag setting
inline void AlphaLCD::isRightToLeft(bool set) {
	set ? rightToLeft() : leftToRight();
}

//! Helper method to set on/off autoscroll
//! \param bool the flag setting
inline void AlphaLCD::isAutoscroll(bool set) {
	set ? autoscroll() : noAutoscroll();
}

/**
 * \brief Create one of the 8 CGRAM memory locations from 0x00 to 0x07
 * with a user defined characters
 *
 * The characters should be 5x8 dots font patterns, so the definition
 * should be an 8 bytes array with the less 5 bits set accordingly with
 * the pixels values of the character pattern.
 *
 * \param int The character memory location
 * \param int the character map 8 bytes array
 */
void AlphaLCD::createChar(uint8_t location, uint8_t charmap[]) {

  location &= 0x7; // No more than 7 memory locations

  command(LCD_SETCGRAMADDR | (location << 3));

  // Character definition loop
  for (int i=0; i<8; i++) {
    write(charmap[i]);
  }
}

/**
 * Helper method to send commands to the device.
 */
inline void AlphaLCD::command(uint8_t value) {
  send(value, false);
}

/**
 * Helper method to send data to the device.
 */
inline size_t AlphaLCD::write(uint8_t value) {
  send(value, true);
}

/**
 * Write either command or data, with automatic 4/8-bit selection
 */
void AlphaLCD::send(uint8_t value, uint8_t mode) {

  if (_displayfunction & LCD_8BITMODE) {
    write8bits(value, mode);
  } else {
    write4bits(value>>4, mode);
    write4bits(value, mode);
  }
}

/**
 * Write the character on the device, 4 bits mode
 *
 * \params int the character data
 * \params mode the backlight on/off writing mode always on by default
 */
void AlphaLCD::write4bits(uint8_t value, uint8_t mode) {
	int EN_SWITCH = B00000010;
	int RS_SWITCH = B00000001;
	int cmd = 0;
	int data = 0;
	if (!mode) {
		cmd = 0 | _backlight;
	} else {
		cmd = LCD_RS_PIN | _backlight;
	}
    data = value<<4 & B11110000;
	cmd |= EN_SWITCH;
	digitalWrite(_latch_pin, HIGH);
	shiftOut(_data_pin, _clock_pin, LSBFIRST, data | cmd);
	digitalWrite(_latch_pin, LOW);
	delayMicroseconds(1);

	cmd &= ~EN_SWITCH;
	digitalWrite(_latch_pin, HIGH);
	shiftOut (_data_pin, _clock_pin, LSBFIRST, data | cmd);
	digitalWrite(_latch_pin, LOW);
	delayMicroseconds(1);

	cmd |= EN_SWITCH;
	digitalWrite(_latch_pin, HIGH);
	shiftOut(_data_pin, _clock_pin, LSBFIRST, data | cmd);
	digitalWrite(_latch_pin, LOW);

	delayMicroseconds(100);
}

/**
 * Write the character on the device, 8 bits mode
 *
 * \params int the character data
 * \params mode the backlight on/off writing mode always on by default
 */
void AlphaLCD::write8bits(uint8_t value, uint8_t mode) {
	int EN_SWITCH = B00000010;
	int RS_SWITCH = B00000001;
	int cmd = 0;
	if (!mode) {
		cmd = 0;
	} else {
		cmd = RS_SWITCH;
	}
	//set enable low
	cmd |= EN_SWITCH;
	digitalWrite (_latch_pin, HIGH);
	shiftOut(_data_pin, _clock_pin, LSBFIRST, cmd);
	shiftOut(_data_pin, _clock_pin, LSBFIRST, value);
	digitalWrite (_latch_pin, LOW);
	//delay (500);
	//set enable high;
	cmd &= ~EN_SWITCH;
	digitalWrite (_latch_pin, HIGH);
	shiftOut(_data_pin, _clock_pin, LSBFIRST, cmd);
	shiftOut(_data_pin, _clock_pin, LSBFIRST, value);
	digitalWrite (_latch_pin, LOW);
	delayMicroseconds (1);
	//delay (500);
	//set enable low
	cmd |= EN_SWITCH;
	digitalWrite (_latch_pin, HIGH);
	shiftOut(_data_pin, _clock_pin, LSBFIRST, cmd);
	shiftOut(_data_pin, _clock_pin, LSBFIRST, value);
	digitalWrite (_latch_pin, LOW);
	delayMicroseconds (100);

}

