/**
 * \file AlphaLCD.h
 * \brief AlphaLCD, allows the LCD to be operated via a shift register
 *
 * \page LCD Alphanumeric Display Controller
 * This class controls the hardware LCD Alphanumeric display. The details on the usage
 * and hardware features are described in the the Display datasheet component data sheet\n
 This library is a variant derived by the original Arduino ShiftLCD library
 with some improvements and more parametrization. For better readability and
 developers customization and usage a doxygen-based comment set has been added.
 *
 * \section copy_sec Credits, Copyright and License
 * \date January 2015
 * \version 2.0
 * \author Created by \a Enrico \a Miglino for Balearic Dynamics Spain\n
 * Documentation revision by \a Maria \a Antonietta \a Ricagno \n
 * Info and contacts: balearicdynamics@gmail.com - http://www.balearicdynamics.com
 *
 * \note This software is licensed as open source under the GNU license.
*/

#ifndef AlphaLCD_h
#define AlphaLCD_h

#include <inttypes.h>
#include "Print.h"

#define LCD_CLEARDISPLAY 0x01		//! Display clear command
#define LCD_RETURNHOME 0x02			//! Display home command
#define LCD_ENTRYMODESET 0x04		//! Display character entry mode (direction) command
#define LCD_DISPLAYCONTROL 0x08		//! Display control (on/off) command
#define LCD_CURSORSHIFT 0x10		//! Display cursor control command
#define LCD_FUNCTIONSET 0x20		//! Display function setting command
#define LCD_SETCGRAMADDR 0x40		//! CGRAM address setting command
#define LCD_SETDDRAMADDR 0x80		//! DDRAM address setting command

#define LCD_ENTRYRIGHT 0x00				//! Bit flag for display characters direction
#define LCD_ENTRYLEFT 0x02				//! Bit flag for display characters direction
#define LCD_ENTRYSHIFTINCREMENT 0x01	//! Bit flag for display characters direction
#define LCD_ENTRYSHIFTDECREMENT 0x00	//! Bit flag for display characters direction

#define LCD_DISPLAYON 0x04		//! Bit flag for display on/off controls
#define LCD_DISPLAYOFF 0x00		//! Bit flag for display on/off controls
#define LCD_CURSORON 0x02		//! Bit flag for display on/off controls
#define LCD_CURSOROFF 0x00		//! Bit flag for display on/off controls
#define LCD_BLINKON 0x01		//! Bit flag for display on/off controls
#define LCD_BLINKOFF 0x00

#define LCD_DISPLAYMOVE 0x08	//! Bit flag for function setting
#define LCD_CURSORMOVE 0x00		//! Bit flag for function setting
#define LCD_MOVERIGHT 0x04		//! Bit flag for function setting
#define LCD_MOVELEFT 0x00		//! Bit flag for function setting

#define LCD_8BITMODE 0x10	//! Bit flag for function setting
#define LCD_4BITMODE 0x00	//! Bit flag for function setting
#define LCD_2LINE 0x08		//! Bit flag for function setting
#define LCD_1LINE 0x00		//! Bit flag for function setting
#define LCD_5x10DOTS 0x04	//! Bit flag for function setting
#define LCD_5x8DOTS 0x00	//! Bit flag for function setting

#define LCD_RS_PIN 0x01	//! pin for functional output
#define LCD_EN_PIN 0x02	//! pin for functional output
#define LCD_BL_PIN 0x04	//! pin for functional output

class AlphaLCD: public Print {
public:

  AlphaLCD();

  AlphaLCD(uint8_t dp, uint8_t cp, uint8_t lp);

  void init(uint8_t dp, uint8_t cp, uint8_t lp);

  void begin(uint8_t cols, uint8_t rows, uint8_t charsize = LCD_5x8DOTS);

  void clear();
  void home();

  void noDisplay();
  void display();
  void noBlink();
  void blink();
  void noCursor();
  void cursor();
  void scrollDisplayLeft();
  void scrollDisplayRight();
  void leftToRight();
  void rightToLeft();
  void autoscroll();
  void noAutoscroll();
  void isDisplay(bool set);
  void isCursor(bool set);
  void isBlinking(bool set);
  void isRightToLeft(bool set);
  void isAutoscroll(bool set);
  void createChar(uint8_t, uint8_t[]);
  void setCursor(uint8_t, uint8_t);
  virtual size_t write(uint8_t);
  void command(uint8_t);
private:
	void send(uint8_t, uint8_t);
	void write4bits(uint8_t, uint8_t);
	void write8bits(uint8_t, uint8_t);

  uint8_t _clock_pin;
  uint8_t _data_pin;
  uint8_t _latch_pin;
  uint8_t _smart_enable;
  uint8_t _displayfunction;
  uint8_t _displaycontrol;
  uint8_t _displaymode;
  uint8_t _backlight;
  uint8_t _initialized;
  uint8_t _numlines,_currline;
};

#endif
